'********************************************************************************************
'
' SpellCheck.inc
'
' AUTHOR: d_homana@yahoo.com.au
' VERSION:  V0.9
' DATE: 21 August 2011
'
'----------------- Main Dictionary functions ----------------------------------------------------
DECLARE FUNCTION InitDictionary() AS integer                        ' Dialog box select
DECLARE FUNCTION FindWord(inpWord AS string) AS integer             ' The main lookup function

'---------------- Other Dictionary functions ----------------------------------------------------
DECLARE FUNCTION IsSpellCheckReady() AS integer
DECLARE FUNCTION DictionaryWordCount() AS integer                   ' Stats on dictionary
DECLARE FUNCTION FilterBothLists() AS integer                       ' Cleanup list remove non-standard chars
DECLARE FUNCTION TestIntegrity() AS integer                         ' Ensure dictionary ok

DECLARE SUB ClearRejectList                                         ' New text file loaded
DECLARE SUB AddSkipWord(NewWord AS string)                          ' Words to ignore (RejectList)
DECLARE FUNCTION CheckRejectList(inpWord AS string) AS integer      ' Skip this word?

'---------------- Crossword Functions -----------------------------------------------------------
DECLARE FUNCTION BuildCrossWordList(NumChars AS integer, Sender AS QStringList) AS integer

'-------------------------------------------------------------------------------------------------
DEFSTR DictionaryFile                                               ' Main dictionary file name
DEFSTR Supplementary                                                ' User added word file (if exists)

DEFINT MainDictionaryLoaded
DEFINT UserDictionaryLoaded
DEFINT SearchReady

DIM TempList AS QStringList                                         ' Used at init and spell-check file

DIM MainWordList AS QStringList                                     ' DictionaryFile list
DIM UserWordList AS QStringList                                     ' Supplementary list

DEFINT MainWordCount
DEFINT UserWordCount
DEFINT MainWordIndex(26)                                            ' Start index of A,B,C...Z words
DEFINT UserWordIndex(26)                                            ' ditto

'----------------------- Main Dictionary Functions -------------------------------------
' Store start index of A,B,C..etc words in main dictionary array
FUNCTION BuildMainIndex() AS integer
DEFINT i, idx
DEFSTR tok1, tok2

    Result=0
    FOR i=1 TO 26: MainWordIndex(i)=-1: NEXT i                       ' Init

    tok2=LEFT$(MainWordList.Item(0),1)
    IF tok2="A" THEN MainWordIndex(1)=0

    idx=1: tok1="A"                                                 ' Starting at "A" words
    FOR i=0 TO MainWordList.ItemCount-1
        tok2=LEFT$(MainWordList.Item(i),1)
        IF tok2<tok1 THEN
            Result=i
            EXIT FUNCTION                                           ' Error?
        END IF
        IF tok2>tok1 THEN
            INC idx
            MainWordIndex(idx)=i
            tok1=tok2
        END IF
    NEXT i
END FUNCTION

FUNCTION BuildUserIndex() AS integer
DEFINT i, idx
DEFBYTE CharVal, PrevVal
DEFSTR token

    Result=0
    FOR i=1 TO 26: UserWordIndex(i)=-1: NEXT i                      ' Init - all invalid

    FOR i=0 TO UserWordList.ItemCount-1
        token=LEFT$(UserWordList.Item(i),1)
        IF (token<"A") OR (token>"Z") THEN
            Result=i
            EXIT FUNCTION
        END IF
        idx=ASC(token)-&H40
        IF UserWordIndex(idx)=-1 THEN
            UserWordIndex(idx)=i
        END IF
    NEXT i
END FUNCTION

FUNCTION DictionaryLoad() AS integer
    Result=False
    MainWordList.Clear
    MainWordList.LoadFromFile(DictionaryFile)                       ' Load main dictionary list
    IF BuildMainIndex()>0 THEN EXIT FUNCTION
    MainDictionaryLoaded=True
    IF (Supplementary<>"") AND FileExists(Supplementary) THEN       ' User list exist?
        UserWordList.Clear
        UserWordList.LoadFromFile(Supplementary)                    ' Add User list
        IF BuildUserIndex()>0 THEN EXIT FUNCTION
        UserDictionaryLoaded=True
    END IF
    ClearRejectList                                                 ' Clear "rejected" list
    SearchReady=True
    Result=True
END FUNCTION

SUB MakeUserFileName
DEFSTR fPath, fName, fType

    fPath="": fName=DictionaryFile                                  ' Do user list if exists
    IF INSTR(fName,"\")>0 THEN
        fPath=LEFT$(fName,RINSTR(fName,"\"))
        fName=fName-fPath                                           ' Name of main file no path
    END IF
    fType=LEFT$(FIELD$(fName,"_",2),2)                              ' 2 chars after "_"
    fPath=fPath+"UserDictionary_"+fType+".txt"                      ' User additional file
    Supplementary=fPath
END SUB

SUB GetMainFileName
DIM fDialog AS QOpenDialog

    fDialog.InitialDir=Application.Path
    fDialog.Caption="Load Dictionary File"
    fDialog.Filter="Dictionary Files|MainDictionary_*.txt"
    IF fDialog.Execute THEN
        DictionaryFile=fDialog.FileName
    END IF
END SUB

FUNCTION InitDictionary() AS integer
DEFINT errors

    Result=0
    MainDictionaryLoaded=False
    UserDictionaryLoaded=False
    DictionaryFile=""
    Supplementary=""                                                ' User added word file (if exists)
    SearchReady=False
    GetMainFileName
    IF DictionaryFile="" THEN EXIT FUNCTION                         ' User cancelled out?
    MakeUserFileName
    IF DictionaryLoad()=False THEN                                  ' Get main (plus user if exist)
        IF MessageBox("Dictionary corrupt! Attempt to fix?","Dictionary Maintenance",&h24)=mrYes THEN
            errors=FilterBothLists()
            IF errors>0 THEN
                MessageBox("Found "+STR$(errors)+" errors. All fixed.","Dictionary Maintenance",&h30)
            END IF
        END IF
    END IF
    Result=MainWordList.ItemCount+UserWordList.ItemCount
END FUNCTION

FUNCTION IsSpellCheckReady() AS integer
    Result=SearchReady
END FUNCTION

FUNCTION DictionaryWordCount() AS integer
    Result=0
    IF SearchReady THEN
        Result=MainWordList.ItemCount+UserWordList.ItemCount
    END IF
END FUNCTION

'---Find word in dictionary---
FUNCTION FindWord(inpWord AS string) AS integer
DEFINT i, start, endit
DEFSTR ThisWord, TestWord, midWord

    Result=False                                                    ' Default fail
    ThisWord=UCASE$(RTRIM$(LTRIM$(inpWord)))

    IF MainWordList.ItemCount<10 THEN                               ' Small list of words?
        FOR i=0 TO MainWordList.ItemCount-1                         ' Just do a straight search
            IF ThisWord=MainWordList.Item(i) THEN
                Result=True
                EXIT FOR
            END IF
        NEXT i
        EXIT FUNCTION
    END IF

    i=ASC(LEFT$(ThisWord,1))-&H40                                   ' Which index?
    start=MainWordIndex(i)                                          ' Start idx of this alpha word
    IF start=-1 THEN EXIT FUNCTION                                  ' No entries? Return False

    IF i=26 THEN                                                    ' "Z"?
        endit=MainWordList.ItemCount-1                              ' END of dictionary list!!
    ELSE
        endit=MainWordIndex(i+1)                                    ' Next index slot
        IF endit=-1 THEN                                            ' IF next letter not present
            endit=MainWordList.ItemCount-1                          ' Just search remainder of list FIXME!!!
        ELSE
            DEC endit
        END IF
    END IF
    IF start<endit THEN
        WHILE (endit-start)>1
            i=start+(endit-start)\2
            midWord=MainWordList.Item(i)
            IF ThisWord=midWord THEN
                Result=True                                         ' Found it!
                EXIT FUNCTION
            END IF
            IF ThisWord>midWord THEN start=i+1 ELSE endit=i-1       ' Test which way one to move
        WEND
        ' NB ONLY two words left to check (endit-start)+1
        IF (ThisWord=MainWordList.Item(start)) OR (ThisWord=MainWordList.Item(endit)) THEN
            Result=True                                             ' Found it!
            EXIT FUNCTION
        END IF
    END IF
    IF UserDictionaryLoaded=FALSE THEN EXIT FUNCTION                ' Exit False

    ' Now check the user additional list
    IF UserWordList.ItemCount<10 THEN                               ' Small list of words?
        FOR i=0 TO UserWordList.ItemCount-1                         ' Just do a straight search
            IF ThisWord=UserWordList.Item(i) THEN
                Result=True
                EXIT FOR
            END IF
        NEXT i
        EXIT FUNCTION
    END IF

    i=ASC(LEFT$(ThisWord,1))-&H40                                   ' Which index?
    start=UserWordIndex(i)                                          ' Start idx of this alpha word
    IF start=-1 THEN EXIT FUNCTION                                  ' No entries? Return False
    endit=UserWordList.ItemCount-1                                  ' END of user list!!

    WHILE (endit-start)>1
        i=start+(endit-start)\2
        midWord=UserWordList.Item(i)
        IF ThisWord=midWord THEN
            Result=True                                         ' Found it!
            EXIT FUNCTION
        END IF
        IF ThisWord>midWord THEN start=i+1 ELSE endit=i-1       ' Test which way one to move
    WEND
    IF (ThisWord=UserWordList.Item(start)) OR (ThisWord=UserWordList.Item(endit)) THEN
        Result=True                                             ' Found it!
        EXIT FUNCTION
    END IF
END FUNCTION
'-------------------------------- Main Functions End ----------------------------------------

' Add a new word to users list and update file & index array
' Creates new file if not exist!
SUB AddNewWord(NewWord AS string)
DEFSTR tstr

    tstr=UCASE$(LTRIM$(RTRIM$(NewWord)))
    UserWordList.AddItems(tstr)
    UserWordList.Sort
    BuildUserIndex
    UserWordList.SaveToFile(Supplementary)
    UserDictionaryLoaded=True
END SUB

FUNCTION IsAlpha(ThisWord AS string) AS integer
DEFINT i
DEFSTR tstr, tok

    Result=True
    i=1
    DO
        tok=tstr[i]
        SELECT CASE tok
            CASE "A" TO "Z", "a" TO "z"
            CASE ELSE
                Result=False
                EXIT DO
        END SELECT
        INC i
    LOOP UNTIL i>LEN(tstr)
END FUNCTION

' Check for "funnies" (non-alpha English) letters in words
FUNCTION FilterBothLists() AS integer
DEFINT i, errors=0, TotalErrors=0
DEFSTR tstr

    TempList.Clear
    FOR i=0 TO MainWordList.ItemCount-1
        tstr=MainWordList.Item(i)
        IF IsAlpha(tstr) THEN
            TempList.AddItems(tstr)
        ELSE
            INC errors
        END IF
    NEXT i
    IF errors>0 THEN
        MainWordList.SaveToFile(LEFT$(DictionaryFile,RINSTR(DictionaryFile,"."))+"err")
        TempList.Sort
        TempList.SaveToFile(DictionaryFile)
        MainWordList.Clear
        MainWordList.AddList(TempList)
        IF BuildMainIndex()<>0 THEN
            Showmessage "Internal ERROR: Unexpected result in FilterBothLists()"
            TempList.Clear
            Result=-1
            EXIT FUNCTION
        END IF
        TotalErrors=errors: errors=0
    END IF
    IF Supplementary<>"" THEN
        TempList.Clear
        FOR i=0 TO UserWordList.ItemCount-1
            tstr=UserWordList.Item(i)
            IF IsAlpha(tstr) THEN
                TempList.AddItems(tstr)
            ELSE
                INC errors
            END IF
        NEXT i
        IF errors>0 THEN
            UserWordList.SaveToFile(LEFT$(Supplementary,RINSTR(Supplementary,"."))+"err")
            TempList.Sort
            TempList.SaveToFile(Supplementary)
            UserWordList.Clear
            UserWordList.AddList(TempList)
            IF BuildUserIndex()<>0 THEN
                Showmessage "Internal ERROR: Unexpected result in FilterBothLists()"
                TempList.Clear
                Result=-1
                EXIT FUNCTION
            END IF
            TotalErrors=TotalErrors+errors
        END IF
    END IF
    TempList.Clear
    Result=TotalErrors
END FUNCTION

'--------------------------Maintenace of dictionary---------------------------------
' Words not in either dictionary, UNSORTED, and only checked once
SUB AddSkipWord(NewWord AS string)
DEFSTR tstr

    tstr=UCASE$(LTRIM$(RTRIM$(NewWord)))                            ' To be safe!
    IF LEN(tstr)>1 THEN
        TempList.AddItems(tstr)
    END IF
END SUB

'---Check word in skip list---
' Shouldn't be too many in this list???? Straight "walking" check
FUNCTION CheckRejectList(inpWord AS string) AS integer
DEFINT i
DEFSTR ThisWord

    Result=False                                                    ' Default fail
    ThisWord=UCASE$(RTRIM$(LTRIM$(inpWord)))
    FOR i=0 TO TempList.ItemCount-1
        IF TempList.Item(i)=ThisWord THEN
            Result=True                                             ' Skip this word
            EXIT FUNCTION
        END IF
    NEXT i
END FUNCTION

SUB ClearRejectList
    TempList.Clear
END SUB


' Check for errors in mainlist
SUB FilterDictionary
DEFINT i, j
DEFSTR tstr, tok

    TempList.Clear
    FOR i=0 TO MainWordList.ItemCount-1
        tstr=MainWordList.Item(i)
        IF LTRIM$(tstr)="" THEN
            TempList.AddItems(STR$(i)+" blank line")
        ELSE
            IF IsAlpha(tstr)=False THEN TempList.AddItems(STR$(i)+" "+tstr)
        END IF
    NEXT i
    IF TempList.ItemCount>0 THEN
        IF MessageBox("Found "+STR$(TempList.ItemCount)+" bad words/lines! Delete?", _
                                                    "Integrity check", &h24)=mrYes THEN
            FOR i=TempList.ItemCount-1 TO 0 STEP -1
                tstr=TempList.Item(i)
                j=VAL(FIELD$(tstr," ",1))
                MainWordList.DelItems(j)
            NEXT i
            MainWordList.SaveToFile(DictionaryFile)
            BuildMainIndex
            TempList.SaveToFile("DeletedList.txt")
        END IF
        TempList.Clear
    END IF
END SUB

' Check the entire dictionary
FUNCTION TestIntegrity() AS integer
DEFINT i, ok
DEFSTR tstr, msg
DIM xList AS QStringList

    Result=False: ok=True
    IF SearchReady=False THEN                                       ' Dictionary not loaded
        ShowMessage "Dictionary not loaded!"
        EXIT FUNCTION
    END IF
    FilterDictionary

    xList.Clear: xList.AddList(MainWordList)
    xList.Sort
    FOR i=0 TO MainWordList.ItemCount-1
        tstr=MainWordList.Item(i)
        IF tstr<>xList.Item(i) THEN
            IF MessageBox("Dictionary needs to be resorted! Proceed?", _
                                                    "Integrity check", &h24)=mrYes THEN
                MainWordList.Clear
                MainWordList.AddList(xList)
                MainWordList.SaveToFile(DictionaryFile)
                BuildMainIndex
                xList.Clear
                ok=True
                EXIT FOR
            ELSE
                ok=False
            END IF
        END IF
    NEXT i
    IF ok=False THEN EXIT FUNCTION                                  ' Ret fail! Don't fix(???)

    IF UserWordList.ItemCount>1 THEN
        xList.Clear: xList.AddList(UserWordList)
        xList.Sort
        FOR i=0 TO UserWordList.ItemCount-1
            tstr=UserWordList.Item(i)
            IF tstr<>xList.Item(i) THEN
                IF MessageBox("Dictionary needs to be resorted! Proceed?", _
                                                        "Integrity check", &h24)=mrYes THEN
                    UserWordList.Clear
                    UserWordList.AddList(xList)
                    UserWordList.SaveToFile(Supplementary)
                    BuildUserIndex
                    xList.Clear
                    Result=True
                    EXIT FUNCTION
                ELSE
                    ok=False
                END IF
            END IF
        NEXT i
    END IF
    Result=ok
END FUNCTION

' Merge UserWordList list into MainWordList
' This is done on closing, so no need to build new index table
' Save main file and delete user file
SUB CombineUserFile
DEFINT i
DEFSTR tstr

    IF SearchReady=False THEN EXIT SUB                                  ' Dictionary not loaded!
    IF UserDictionaryLoaded=False THEN EXIT SUB                         ' Nothing to combine!

    tstr=LEFT$(DictionaryFile,RINSTR(DictionaryFile,"."))+"bak"
    MainWordList.SaveToFile(tstr)                                       ' Backup just in case
    i=MainWordList.ItemCount
    i=i+UserWordList.ItemCount
    MainWordList.AddList(UserWordList)
    IF MainWordList.ItemCount=i THEN
        MainWordList.Sort
        MainWordList.SaveToFile(DictionaryFile)
        KILL Supplementary
        DoEvents
    ELSE
        MainWordList.LoadFromFile(DictionaryFile)
        ShowMessage "Failed combine!"
    END IF
END SUB

'----------------------------- Crossword stuff ----------------------------------------------
' For speed, build a list of 'n' char words in supplied QStringList
FUNCTION BuildCrossWordList(NumChars AS integer, Sender AS QStringList) AS integer
DEFINT i
DEFSTR tstr

    Result=0
    IF SearchReady=False THEN EXIT FUNCTION
    Sender.Clear
    FOR i=0 TO MainWordList.ItemCount-1
        tstr=MainWordList.Item(i)
        IF LEN(tstr)=NumChars THEN Sender.AddItems(tstr)
    NEXT i
    IF UserDictionaryLoaded THEN
        FOR i=0 TO UserWordList.ItemCount-1
            tstr=UserWordList.Item(i)
            IF LEN(tstr)=NumChars THEN Sender.AddItems(tstr)
        NEXT i
    END IF
    Result=Sender.ItemCount
END FUNCTION

'********************************************************************************************

