'
' WordPlay.bas
'
' A bit more than a spell checker!!
'
' AUTHOR:   d_homans@yahoo.com.au
' VERSION:  V0.9
' DATE:     21 August 2011
'
'   PROGRAM:
'       Uses ONLY Rapidq code with QStringLists to sort, display etc
'       Uses a QStringGrid to display text files & search data (slow)
'       Create/update new lists of words, update dictionary and more
'
'       Main dictionary file "MainDictionary_EN.txt"
'       User additions file "UserDictionary_EN.txt"
'
'   DICTIONARY FILE:  EOWL ~129,000 words (lower case) - Thanks to Jacques for info
'       http://dreamsteep.com/projects/98-software/53-the-english-open-word-list-eowl.html
'
'       Unzipped all, joined and UCASEd, LF->CRLFd then renamed "MainDictionary_EN.txt"
'       This file has 280+ words with non-English characters in it: eg NGSTRM, BTISE
'       The list is limited to 10 chars/word max & NO hyphenated words by the provider.
'       Final word count 128,771 words.
'       This demo removes that limit and allow any sized word to be added.
'
$APPTYPE GUI
$TYPECHECK ON
$OPTION ICON "a-z.ico"

CONST VERS="V0.9"

$DEFINE TIME_ME                                                     ' Show some timings?

$INCLUDE <Rapidq2.inc>
$INCLUDE "SpellCheck.inc"

$RESOURCE info AS "info.txt"

CONST CRLF=CHR$(13)+CHR$(10)

$DEFINE MAX_LETRS    12

CONST KILL_PROGBAR = -1                                             ' Kill progress bar window
$DEFINE PBM_SETRANGE        &h401
$DEFINE PBM_SETPOS          &h402
$DEFINE PBM_DELTAPOS        &h403
$DEFINE PBM_SETSTEP         &h404
$DEFINE PBM_STEPIT          &h405
$DEFINE PBM_SETRANGE32      &h406
$DEFINE PBM_GETRANGE        &h407
$DEFINE PBM_GETPOS          &h408
$DEFINE PBM_SETBARCOLOR     &h409
$DEFINE PBM_SETBKCOLOR      &h2001                                  ' ProgressBar background colour

DECLARE FUNCTION UpdateWindow LIB "user32" ALIAS "UpdateWindow" (hWnd AS long) AS integer
DECLARE FUNCTION DestroyWindow LIB "user32" ALIAS "DestroyWindow" (hwnd AS long) AS long

'---------------- Other functions ---------------------------------------------------------------
DECLARE FUNCTION InitProgressBar(hWnd AS long, Left AS integer, Top AS integer, Width AS integer, _
                                                                    Height AS integer) AS integer
DECLARE SUB UpdateGauge(barVal AS integer)                          ' Progress bar stuff

$IFDEF TIME_ME
DECLARE FUNCTION ElapsedTimer(tStart AS single) AS single           ' tStart=0 ret Timer else ret tStart-Timer
$ENDIF
'---------------- Main functions ---------------------------------------------------------------
DECLARE SUB Initialise
DECLARE SUB LoadDictionary
DECLARE SUB LoadTextFile
DECLARE SUB SaveFile
DECLARE SUB SpellCheckFile
DECLARE SUB FileStats
DECLARE SUB ConvertUpper
DECLARE SUB ConvertLower
DECLARE SUB ConvertTabs
DECLARE SUB SortAZ
DECLARE SUB FilterTextFile
DECLARE SUB CreateDictionary
DECLARE SUB AddToDictionary
DECLARE SUB CompactDictionary
DECLARE SUB SpeedTest
DECLARE SUB TestDictionary
DECLARE SUB DisplayInfo
DECLARE SUB ReloadTextFile
DECLARE SUB About
DECLARE SUB Done

DECLARE SUB CrossWordInit
DECLARE SUB CrosswordStartSearch
DECLARE SUB CrosswordChangeCount

DEFSTR TestFile=""                                                  ' User selected file
DEFINT TestFileLoaded
DEFINT KeepSearching                                                ' Search abort flag
DEFINT GotProgressBar
DIM ResultList AS QStringList
DEFLNG hParent, hProgBar                                            ' Progress bar

DIM xEdit(1 TO MAX_LETRS) AS QEDIT

CREATE Form AS QFORM
    Width = 800
    Height = Screen.Height*3\4
    BorderStyle = bsSingle
    Caption = "WordPlay "+VERS
    DelBorderIcons(biMinimize,biMaximize)
    Center
    CREATE MainMenu AS QMAINMENU
        CREATE AMenu AS QMENUITEM
            Caption = "File"
            CREATE A1 AS QMENUITEM
                Caption = "Load Dictionary"
                OnClick = LoadDictionary
            END CREATE
            CREATE A2 AS QMENUITEM
                Caption = "Load Text File"
                OnClick = LoadTextFile
            END CREATE
            CREATE A3 AS QMENUITEM
                Caption = "Save File"
                Enabled = False
                OnClick = SaveFile
            END CREATE
            CREATE mnuSep1 AS QMENUITEM
                Caption = "-"
            END CREATE
            CREATE A4 AS QMENUITEM
                Caption = "Exit"
                OnClick = Done
            END CREATE
        END CREATE
        CREATE BMenu AS QMENUITEM
            Caption = "Utilities"
            CREATE B1 AS QMENUITEM
                Caption = "SpellCheck File"
                Enabled = False
                OnClick = SpellCheckFile
            END CREATE
            CREATE mnuSep2 AS QMENUITEM
                Caption = "-"
            END CREATE
            CREATE B2 AS QMENUITEM
                Caption = "Add to Dictionary"
                OnClick = AddToDictionary
                Enabled = False
            END CREATE
            CREATE B3 AS QMENUITEM
                Caption = "Create Dictionary"
                OnClick = CreateDictionary
                Enabled = False
            END CREATE
            CREATE mnuSep3 AS QMENUITEM
                Caption = "-"
            END CREATE
            CREATE B4 AS QMENUITEM
                Caption = "CrossWord Lookup"
                Enabled = False
                OnClick = CrossWordInit
            END CREATE
        END CREATE
        CREATE CMenu AS QMENUITEM
            Caption = "Text Functions"
            CREATE C1 AS QMENUITEM
                Caption = "File Stats"
                Enabled = False
                OnClick = FileStats
            END CREATE
            CREATE C2 AS QMENUITEM
                Caption = "Convert to UCASE"
                Enabled = False
                OnClick = ConvertUpper
            END CREATE
            CREATE C3 AS QMENUITEM
                Caption = "Convert to LCASE"
                Enabled = False
                OnClick = ConvertLower
            END CREATE
            CREATE C4 AS QMENUITEM
                Caption = "TABs -> 4sp"
                Enabled = False
                OnClick = ConvertTabs
            END CREATE
            CREATE C5 AS QMENUITEM
                Caption = "Sort A-Z"
                Enabled = False
                OnClick = SortAZ
            END CREATE
            CREATE mnuSep4 AS QMENUITEM
                Caption = "-"
            END CREATE
            CREATE C6 AS QMENUITEM
                Caption = "Filter"
                Enabled = False
                OnClick = FilterTextFile
            END CREATE
            CREATE mnuSep5 AS QMENUITEM
                Caption = "-"
            END CREATE
            CREATE C7 AS QMENUITEM
                Caption = "Reload File"
                Enabled = False
                OnClick = ReloadTextFile
            END CREATE
        END CREATE
        CREATE DMenu AS QMENUITEM
            Caption = "Maintenance"
            CREATE D1 AS QMENUITEM
                Caption = "Compact Dictionary"
                Enabled = False
                OnClick = CompactDictionary
            END CREATE
            CREATE D2 AS QMENUITEM
                Caption = "Integrity Test"
                Enabled = False
                OnClick = TestDictionary
            END CREATE
            CREATE D3 AS QMENUITEM
                Caption = "Speed Test"
                Enabled = False
                OnClick = SpeedTest
            END CREATE
        END CREATE
        CREATE EMenu AS QMENUITEM
            Caption = "Help"
            CREATE E1 AS QMENUITEM
                Caption = "WordPlay Information"
                OnClick = DisplayInfo
            END CREATE
            CREATE E2 AS QMENUITEM
                Caption = "About"
                OnClick = About
            END CREATE
            CREATE E3 AS QMENUITEM
                Caption = "Reload File"
                Enabled = False
                OnClick = ReloadTextFile
            END CREATE
        END CREATE
    END CREATE
    CREATE stat AS QSTATUSBAR
        SizeGrip=False
        SimplePanel=True
    END CREATE
    CREATE TextGrid AS QStringGrid
        Left = 10
        Top = 0
        Width = Form.ClientWidth-20
        Height = Form.ClientHeight-stat.Height-16
        FixedCols=0
        FixedRows=0
        ColCount=1
        DefaultColWidth=TextGrid.Width-40
        DefaultRowHeight=16
        GridLineWidth=0
        ColumnStyle(0)=gcsNone
        Font.Name="Courier"
    END CREATE
    CREATE xPanel AS QPanel
        Width = 360
        Height = 100
        Left=(Form.Width-xPanel.Width)\2
        Top=(Form.Height-xPanel.Height)\2
        BorderStyle = 1
        Visible = False
        CREATE xBox AS QGroupBox
            Caption = ""
            Left = 10
            Top = 10
            Width = xPanel.Width-20
            Height = 60
            CREATE xLabel AS QLABEL
                Top = 25
                Width = 18
                Height = 18
                Caption = "12"
            END CREATE
            CREATE xScroll AS QSCROLLBAR
                Top = 15
                Width = 35
                Kind = sbVertical
                Min = 1
                Max = 11
                OnChange = CrosswordChangeCount
            END CREATE
        END CREATE
        CREATE xButn AS QBUTTON
            Width = 55
            Height = 20
            Left = (xPanel.Width-xButn.Width)\2
            Top = 73
            Caption = "GO"
            OnClick = CrosswordStartSearch
        END CREATE
    END CREATE
END CREATE

Initialise

Form.ShowModal

SUB EnableUtils
    A3.Enabled=True: B3.Enabled=True
    IF IsSpellCheckReady() THEN
        B1.Enabled=True: B2.Enabled=True
        B3.Enabled=False: B4.Enabled=True
    END IF
    C1.Enabled=True: C2.Enabled=True: C3.Enabled=True
    C4.Enabled=True: C5.Enabled=True: C6.Enabled=True
    C7.Enabled=True
END SUB

SUB DisableUtils
    A3.Enabled=False: B1.Enabled=False
    B2.Enabled=False: B3.Enabled=False
    C1.Enabled=False: C2.Enabled=False
    C3.Enabled=False: C4.Enabled=False
    C5.Enabled=False: C6.Enabled=False
    C7.Enabled=False
END SUB

SUB ColorWord(LineNum AS integer, Posn AS integer, strWord AS string)
DEFSTR tstr
DEFINT x, y

    TextGrid.Repaint
    tstr=LEFT$(TextGrid.Cell(0,LineNum),Posn-1)
    x=TextGrid.TextWidth(tstr)+2
    y=(LineNum-TextGrid.TopRow)*TextGrid.DefaultRowHeight+2
    TextGrid.TextOut(x,y,strWord,&h7f,&h7fffff)
END SUB

' If not in the dictionary and not in the skip list, ask what to do
SUB SearchDictionary(LineNum AS integer, Posn AS integer, strWord AS string)
DEFINT res
DEFSTR tstr=strWord

    ColorWord(LineNum,Posn,tstr)
    IF (FindWord(tstr)=False) THEN
        IF (CheckRejectList(tstr)=False) THEN
            ' <<<<<<<<<<<<<<<<Word spelt incorrectly - add code for user to fix????
            res=MessageBox("Save NEW word in the dictionary?","Dictionary Maintenance",&h123)
            SELECT CASE res
                CASE mrYes: AddNewWord(tstr)
                CASE mrNo: AddSkipWord(tstr)
                CASE ELSE
                    KeepSearching=False
            END SELECT
        END IF
    END IF
END SUB

SUB LoadDictionary
DEFINT WordCount
DEFSTR tstr
DEFSNG Chrono

    stat.SimpleText="Loading Dictionary..."
$IFDEF TIME_ME
    Chrono=ElapsedTimer(0)
$ENDIF
    WordCount=InitDictionary()                                      ' Select a dictionary
$IFDEF TIME_ME
    Chrono=ElapsedTimer(Chrono)
$ENDIF
    IF WordCount>0 THEN
        B3.Enabled=False: B4.Enabled=True
        D1.Enabled=True: D2.Enabled=True: D3.Enabled=True
        IF TestFileLoaded THEN B1.Enabled=True: B2.Enabled=True 
        tstr="Dictionary loaded ok. "+STR$(WordCount)+" words in dictionary"
    ELSE
        tstr="ERROR: Dictionary FAILED to load"
    END IF
$IFDEF TIME_ME
    tstr="[Time: "+STR$(Chrono)+"s] "+tstr
$ENDIF
    stat.SimpleText=tstr
END SUB

WITH TextGrid

SUB ClearTextGrid
DEFINT i

    IF .RowCount>0 THEN
        FOR i=0 TO .RowCount-1: .Cell(0,i)="": NEXT i
    END IF
END SUB

FUNCTION GetWordCount() AS integer
DEFINT i, idx, cnt=0
DEFSTR tstr, tok, tWord

    FOR i=0 TO .RowCount-1
        tstr=REPLACESUBSTR$(.Cell(0,i),CHR$(9)," ")
        tWord=""
        FOR idx=1 TO LEN(tstr)
            tok=tstr[idx]
            IF tok=" " THEN
                IF tWord<>"" THEN INC cnt: tWord=""
            ELSE
                tWord=tWord+tok
            END IF
        NEXT idx
        IF tWord<>"" THEN INC cnt
    NEXT i
    Result=cnt
END FUNCTION

SUB FillTextGrid(FileName AS string)
DEFINT i,j
DEFSTR tstr
DIM fList AS QStringList

    .RowCount=.VisibleRowCount
    ClearTextGrid

    fList.LoadFromFile(FileName)
    i=fList.ItemCount
    .RowCount=i
    FOR i=0 TO .RowCount-1
        tstr=fList.Item(i)
        tstr=REPLACESUBSTR$(tstr,CHR$(13),"")
        .Cell(0,i)=tstr
    NEXT i
    .InsertRow(0)                                                   ' Fudge: probs painting on top row(!!)
    .TopRow=1
    fList.Clear
    j=GetWordCount()
    tstr=TestFile+" loaded ["+STR$(i)+" Lines "+STR$(j)+" text tokens]"
    stat.SimpleText=tstr
END SUB

SUB LoadTextFile
DIM fDialog AS QOpenDialog

    IF GotProgressBar>0 THEN UpdateGauge(0)
    stat.SimpleText="Loading Test file..."
    fDialog.Caption="Load Test File"
    fDialog.InitialDir=Application.Path
    fDialog.Filter="Text Files|*.txt|All Files|*.*"
    IF fDialog.Execute THEN
        TestFile=fDialog.FileName
        FillTextGrid(TestFile)
        TestFileLoaded=True
        EnableUtils
        IF IsSpellCheckReady() THEN ClearRejectList
    ELSE
        TestFile="": TestFileLoaded=False
        DisableUtils
    END IF

END SUB

SUB SaveFile
DIM fDialog AS QSaveDialog

    IF .RowCount>1 THEN
        stat.SimpleText="Saving displayed text..."
        fDialog.Caption="Save to File"
        fDialog.InitialDir=Application.Path
        fDialog.Filter="Text Files|*.txt|All Files|*.*"
        IF fDialog.Execute THEN
            stat.SimpleText="File saved"
        ELSE
            stat.SimpleText="Save to File cancelled!"
        END IF
    ELSE
        stat.SimpleText="Nothing to Save!"
    END IF
END SUB

' Parse each line of TestFile for words to check
SUB SpellCheckFile
DEFINT i, endit, CharPos, bar, incr, cntr, LineNum, badchar
DEFSTR tstr, tok, chkword
DEFSNG Chrono

    IF TestFileLoaded=False THEN
        stat.SimpleText="WHAT? Need a file to Spell Check!"
        EXIT SUB
    END IF
    IF GotProgressBar THEN
        bar=0: cntr=0
        incr=.RowCount\100                                          ' < 100 incr=0 ok
        UpdateGauge(0)
    END IF

    stat.SimpleText="Spell Check started..."
    LineNum=0: .TopRow=1: .Repaint
    KeepSearching=True

$IFDEF TIME_ME
    Chrono=ElapsedTimer(0)
$ENDIF
    DO
        tstr=RTRIM$(.Cell(0,LineNum))
        IF tstr<>"" THEN
            i=1: endit=LEN(tstr): CharPos=1: chkword="": badchar=False
            DO
                tok=tstr[i]
                SELECT CASE tok
                    CASE IS > "~"
                        badchar=True
                        EXIT DO
                    CASE " "
                        IF (LEN(chkword)>1) AND (badchar=False) THEN
                            SearchDictionary(LineNum, CharPos, chkword)
                        END IF
                        badchar=False: chkword=""
                    CASE ELSE
                        chkword=chkword+tok
                        IF LEN(chkword)=1 THEN CharPos=i            ' Mark start of a word
                        IF (UCASE$(tok)<"A") OR (UCASE$(tok)>"Z") THEN badchar=True
                END SELECT
                IF KeepSearching=False THEN EXIT DO                 ' User cancel?
                INC i
            LOOP UNTIL i>endit
            IF (LEN(chkword)>1) AND (badchar=False) THEN _
                                SearchDictionary(LineNum, CharPos, chkword)
            badchar=False: chkword=""
        END IF
        IF GotProgressBar THEN
            IF incr=0 THEN                                          ' < 100 lines?
                bar=bar+100\.RowCount
            ELSE                                                    ' > 100 lines
                INC cntr
                IF cntr>=incr THEN
                    cntr=0: INC bar
                END IF
            END IF
            UpdateGauge(bar)
        END IF
        IF KeepSearching=False THEN EXIT DO                         ' User cancel?
        INC LineNum
        IF LineNum>4 THEN .TopRow=.TopRow+1                         ' Start scrolling grid window
    LOOP UNTIL LineNum>.RowCount-1
    .Repaint
    IF GotProgressBar THEN UpdateGauge(0)

    IF KeepSearching=False THEN
        tstr="User cancelled!"
    ELSE
        KeepSearching=False
        tstr="Finished!"
    END IF
'    TempList.SaveToFile("Failures.txt")                             ' Check what was skipped
$IFDEF TIME_ME
    Chrono=ElapsedTimer(Chrono)
    tstr="[Time: "+STR$(Chrono)+"s] "+tstr
$ENDIF
    stat.SimpleText=tstr
    ShowMessage tstr
END SUB

'------------------------------ Utilities ------------------------------------
SUB FileStats
DEFINT i

    IF TestFileLoaded THEN
        i=GetWordCount()
        stat.SimpleText="["+TestFile+"] Lines: "+STR$(.RowCount-1)+" Text tokens: "+STR$(i)
    ELSE
        stat.SimpleText="WHAT? No file loaded to check!"
    END IF
END SUB

SUB ConvertUpper
DEFINT i

    FOR i=0 TO .RowCount-1
        .Cell(0,i)=UCASE$(.Cell(0,i))
    NEXT i
END SUB

SUB ConvertLower
DEFINT i

    FOR i=0 TO .RowCount-1
        .Cell(0,i)=LCASE$(.Cell(0,i))
    NEXT i
END SUB

SUB ConvertTabs
DEFINT i

    FOR i=0 TO .RowCount-1
        .Cell(0,i)=REPLACESUBSTR$(.Cell(0,i),CHR$(9),"    ")
    NEXT i
END SUB

SUB SortAZ
DEFINT i
DEFSTR tstr

    ResultList.Sorted=True
    ResultList.Clear
    FOR i=0 TO .RowCount-1
        tstr=LTRIM$(REPLACESUBSTR$(.Cell(0,i),CHR$(9),""))
        IF tstr<>"" THEN ResultList.AddItems tstr
        .Cell(0,i)=""
    NEXT i
    FOR i=0 TO ResultList.ItemCount-1
        .Cell(0,i)=ResultList.Item(i)
    NEXT i
    .Repaint
    ResultList.Clear
END SUB


FUNCTION IsWordOk(inpWord AS string) AS integer
    Result=False
    IF INSTR(inpWord," ")>0 THEN EXIT FUNCTION
    IF (INSTR(inpWord,"A")=0) AND (INSTR(inpWord,"E")=0) _
        AND (INSTR(inpWord,"I")=0) AND (INSTR(inpWord,"O")=0) _
        AND (INSTR(inpWord,"U")=0) AND (INSTR(inpWord,"Y")=0) THEN
        EXIT FUNCTION
    END IF
    Result=True
END FUNCTION

' Filter the displayed text (from the file)
SUB FilterTextFile
DEFINT i, j
DEFSTR tstr, tok, nWord

    ResultList.Sorted=True
    ConvertTabs
    ConvertUpper
    UpdateGauge(0)
    FOR i=0 TO .RowCount-1
        tstr=.Cell(0,i)
        nWord=""
        FOR j=1 TO LEN(tstr)
            tok=tstr[j]
            IF tok=" " THEN
                nWord=LTRIM$(RTRIM$(nWord))
                IF LEN(nWord)>1 THEN
                    IF IsWordOK(nWord) THEN ResultList.AddItems(nWord)
                END IF
                nWord=""
            ELSEIF tok>"~" THEN
                nWord=""
            ELSEIF (tok<"A") OR (tok>"Z") THEN
                nWord=nWord+" "
            ELSE
                nWord=nWord+tok
            END IF
        NEXT j
        nWord=LTRIM$(RTRIM$(nWord))                             ' Catch end of line
        IF (LEN(nWord)>1) AND (INSTR(nWord," ")=0) THEN
            ResultList.AddItems(nWord)
        END IF
        UpdateGauge(i*100\.RowCount)
    NEXT i
    UpdateGauge(0)
END SUB

SUB AddToDictionary
DEFINT old

    IF (SearchReady()=True) AND (.RowCount>0) THEN
        ResultList.Clear
        ResultList.LoadFromFile(DictionaryFile)
        old=ResultList.ItemCount
        FilterTextFile
        ResultList.SaveToFile(DictionaryFile)
        ClearTextGrid
        stat.SimpleText="File merged with Dictionary ("+ _
                    STR$(ResultList.ItemCount-old)+" words added!)"
        MessageBox("Load new dictionary, to continue","Build Dictionary",&h40)
    END IF
END SUB

FUNCTION SaveTextDictionary(fName AS string) AS integer
    Result=False
    ResultList.Clear
    FilterTextFile
    IF ResultList.ItemCount>0 THEN
        ResultList.SaveToFile(fName)
        ResultList.Clear
        ClearTextGrid
        B3.Enabled=False
        Result=True
    END IF
END FUNCTION

SUB BackupDictionary(fName AS string)
DEFSTR tstr

    tstr=LEFT$(fName,LEN(fName)-3)+"bak"
    RENAME(fName,tstr)
    DoEvents
END SUB

' NB: Uses QOpenDialog for custom Replace with backup!
SUB CreateDictionary
DIM fDialog AS QOpenDialog
DEFINT res=0

    fDialog.InitialDir=Application.Path
    fDialog.Caption="Create Dictionary File"
    fDialog.Filter="Dictionary Files|MainDictionary_*.txt"
    fDialog.FileName="MainDictionary_EN.txt"
    IF fDialog.Execute THEN
        IF FileExists(fDialog.FileName)>0 THEN
            res=MessageBox("Dictionary exists, backup before over-writing?","Create Dictionary",&h123)
            SELECT CASE res
                CASE mrYes: BackupDictionary(fDialog.FileName)
                CASE mrNo: ' Fall through!
                CASE ELSE
                    stat.SimpleText="Create dictionary aborted!"
                    EXIT SUB
            END SELECT
        END IF
        IF SaveTextDictionary(fDialog.FileName) THEN
            stat.SimpleText="New dictionary created!"
            MessageBox("Load new dictionary, to continue","Build Dictionary",&h40)
        ELSE
            stat.SimpleText="ERROR: NO dictionary created!"
        END IF
    ELSE
        stat.SimpleText="Create dictionary cancelled!"
    END IF
    DoEvents
END SUB

'------------------------------ Maintenance ------------------------------------
SUB CompactDictionary
    IF SearchReady THEN
        CombineUserFile
        stat.SimpleText="INFO: Dictionary compacted. "+ _
                STR$(MainWordList.ItemCount)+" words in dictionary"
    ELSE
        stat.SimpleText="INFO: Load dictionary first!"
    END IF
END SUB

SUB TestDictionary
DEFINT i,j

    IF SearchReady THEN
        stat.SimpleText="Start: Dictionary Integrity check...."
        i=DictionaryWordCount()
        IF TestIntegrity() THEN
            j=DictionaryWordCount()
            IF i<>j THEN
                stat.SimpleText="Dictionary Updated ["+STR$(i)+" -> "+STR$(j)+" words]"
            ELSE
                stat.SimpleText="Dictionary secure"
            END IF
        ELSE
            stat.SimpleText="Dictionary bad!"
        END IF
    ELSE
        stat.SimpleText="INFO: Load dictionary first!"
    END IF
END SUB

' Test the dictionary file words against the Dictionary
SUB SpeedTest
DEFINT idx, cntr, toggle, fail
DEFSTR tstr
DEFSNG Chrono
DIM TestMeList AS QStringList

    IF SearchReady=False THEN
        stat.SimpleText="INFO: Load dictionary first!"
        EXIT SUB
    END IF
    stat.SimpleText="Start: FULL dictionary scan....(may take a while)"
    IF GotProgressBar>0 THEN UpdateGauge(0)
    Sleep.MS 250
    TestMeList.Clear
    TestMeList.LoadFromFile(DictionaryFile)
    cntr=0: toggle=0: idx=TestMeList.ItemCount-1
    DoEvents
    Chrono=ElapsedTimer(0)
    DO
        tstr=TestMeList.Item(idx)
        IF FindWord(tstr)=False THEN EXIT DO
        INC cntr
        IF cntr>1000 THEN
            cntr=0
            toggle=1-toggle
            IF toggle THEN stat.SimpleText="*" ELSE stat.SimpleText=""
        END IF
        DEC idx
    LOOP UNTIL idx=0
    Chrono=ElapsedTimer(Chrono)
    DoEvents

    tstr="[Time: "+STR$(Chrono)+"s] "
    IF idx>0 THEN                                                   ' An error?
        tstr=tstr+"Fail! "+STR$(idx)+": "+TestMeList.Item(idx)
    ELSE
        tstr=tstr+"Finished: Word count="+STR$(TestMeList.ItemCount)
    END IF
    TestMeList.Clear
    stat.SimpleText=tstr
    DoEvents
END SUB

SUB Initialise
DEFINT i

    SetWindowLong(Form.Handle, GWL_HWNDPARENT, 0)
    SetWindowLong(Application.Handle, GWL_HWNDPARENT, Form.Handle)
    GotProgressBar=InitProgressBar(Form.Handle,TextGrid.Left, _
                        TextGrid.Top+TextGrid.Height,TextGrid.Width,16)

    IF GotProgressBar>0 THEN UpdateGauge(0)
    FOR i=1 TO MAX_LETRS
        xEdit(i).Parent=xBox
        IF i>6 THEN xEdit(i).Enabled=False ELSE xEdit(i).Enabled=True
        xEdit(i).Left=i*20+10: xEdit(i).Top=20
        xEdit(i).Width=18: xEdit(i).Height=18
        xEdit(i).MaxLength=1: xEdit(i).Text = "?"
        xEdit(i).CharCase=1                                         ' Ucase
        xEdit(i).InputMask = "l"                                    ' A-Z or <space>
    NEXT i
    INC i: xLabel.Left=i*20+10
    INC i: xScroll.Left=i*20+10
    xScroll.Position=7
END SUB

'---------------------------- Crossword search --------------------------
'---Find word in crossword list---FAST!
' Builds a list of 'n' char words first, then straight pattern match
FUNCTION CrosswordSearch(SearchMask AS string) AS integer
DEFINT i, j, cnt, max, res
DEFSTR tstr, tok1, tok2
DIM xList AS QStringList
DEFSNG Chrono

    Result=0
    stat.SimpleText="Searching...."

    Chrono=ElapsedTimer(0)
    ResultList.Clear: xList.Clear: xList.Sorted=True
    cnt=LEN(SearchMask)
    max=BuildCrossWordList(cnt,xList)                               ' Build the list (max words)
    IF max=0 THEN
        stat.SimpleText="NO words found!"
        EXIT FUNCTION
    END IF

    FOR i=0 TO max-1                                                ' Start the pattern match
        tstr=xList.Item(i)
        res=True
        FOR j=1 TO cnt                                              ' Loop each letter
            tok1=SearchMask[j]
            tok2=tstr[j]
            IF (tok1<>"*") AND (tok1<>tok2) THEN                    ' No match?
                res=False
                EXIT FOR
            END IF
        NEXT j
        IF res=True THEN ResultList.AddItems(tstr)                  ' Found one!
    NEXT i
    Chrono=ElapsedTimer(Chrono)
    tstr="[Time: "+STR$(Chrono)+"s] Search completed! Found "+STR$(ResultList.ItemCount)+" words "
    tstr=tstr+"(out of "+STR$(max)+" words of "+STR$(cnt)+" characters)"
    stat.SimpleText=tstr

    IF GotProgressBar>0 THEN UpdateGauge(0)
    FOR i=0 TO ResultList.ItemCount-1                     
        .Cell(0,.Row)=ResultList.Item(i)                            ' Update the screen
        .Row=.Row+1
        IF .Row=.RowCount-1 THEN
            .RowCount=.RowCount+1
        END IF
        IF GotProgressBar>0 THEN UpdateGauge(i*100\ResultList.ItemCount)
    NEXT i
    UpdateGauge(100)
    xList.Clear
    Result=ResultList.ItemCount
END FUNCTION

SUB CrossWordInit
DEFINT i
DEFSTR tstr, tok

    IF SearchReady=False THEN EXIT SUB                              ' Dictionary loaded?
    IF GotProgressBar>0 THEN UpdateGauge(0)
    xButn.Tag=1                                                     ' Go-stop flag
    xPanel.Visible=True                                             ' Get user input
    WHILE xButn.Tag=1
        DoEvents                                                    ' Wait for "GO" butn
    WEND

    ' Leave previous display "as is" until AFTER GO butn clicked!
    ResultList.Clear
    ClearTextGrid
    .RowCount=.VisibleRowCount: .Row=0
    tstr=""
    FOR i=1 TO MAX_LETRS
        IF xEdit(i).Enabled=False THEN EXIT FOR
        tok=UCASE$(xEdit(i).Text)
        IF (tok<"A") OR (tok>"Z") THEN tstr=tstr+"*" ELSE tstr=tstr+tok
    NEXT i
    xPanel.Visible=False
    CrosswordSearch(tstr)
END SUB

SUB CrosswordStartSearch
    xButn.Tag=0                                                     ' GO butn clicked
END SUB

SUB CrosswordChangeCount
DEFINT i, j=MAX_LETRS+1-xScroll.Position                            ' Up arrow = increment count

    xLabel.Caption=" "+STR$(j)
    IF j>9 THEN xLabel.Caption=LTRIM$(xLabel.Caption)

    FOR i=1 TO j: xEdit(i).Enabled=True: NEXT i                     ' Set the input boxes
    IF i<MAX_LETRS+1 THEN
        FOR i=i TO MAX_LETRS: xEdit(i).Enabled=False: xEdit(i).Text="?": NEXT i
    END IF
END SUB

'------------------------------ Info ------------------------------------
SUB DisplayInfo
DIM mem AS QMemoryStream

    DisableUtils
    IF TestFileLoaded THEN E3.Enabled=True
    IF GotProgressBar>0 THEN UpdateGauge(0)
    ClearTextGrid
    mem.ExtractRes(Resource(0))
    mem.Position=0
    .RowCount=mem.lineCount+1
    .LoadFromStream(mem,1,0,mem.LineCount)
    mem.Close
END SUB

SUB ReloadTextFile
DEFINT i

    E3.Enabled=False
    IF TestFileLoaded THEN
        FillTextGrid(TestFile)
        EnableUtils
        ClearRejectList                                             ' Spellcheck clear
    ELSE
        ClearTextGrid
        .Row=0
    END IF
END SUB

END WITH

SUB About
DEFSTR tstr

    tstr="WordPlay "+VERS+CRLF+CRLF
    tstr=tstr+"A spell-checker & crossword search"+CRLF
    tstr=tstr+"demo using pure Rapidq code only!"+CRLF+CRLF
    tstr=tstr+"Author: d_homans@yahoo.com.au"
    MessageBox(tstr,"About",&H40)
END SUB

' On exit, if user words list > 100 suggest compaction
SUB Done
    IF GotProgressBar THEN UpdateGauge(KILL_PROGBAR)
    IF UserWordList.ItemCount>1000 THEN
        IF MessageBox("Compact Dictionary File?","Dictionary Maintenance",&h124)=mrYes THEN
            CombineUserFile
        END IF
    END IF
    Application.Terminate
END SUB

'----------------------------- Additional stuff ----------------------------------------------------
FUNCTION InitProgressBar(hWnd,Left,Top,Width,Height)
DEFLNG PBRange=&H10000*100                                  ' Lo word 0 (min), Hi word (100) max

    Result=False
    hParent=hWnd                                                    ' Save in a global
    hProgBar = CreateWindowEx(0, "msctls_progress32", "", WS_CHILD OR WS_VISIBLE, Left, Top, _
                                                Width, Height, hWnd, 0, Application.hInstance, 0)
    IF hProgBar=0 THEN EXIT FUNCTION

    SendMessage(hProgBar, PBM_SETBKCOLOR, 0, &hffffff)              ' Set progress bar backcolor
    SendMessage(hProgBar, PBM_SETBARCOLOR, 0, &h7f00)               ' Set progress bar forecolor
    SendMessage(hProgBar, PBM_SETRANGE, 0, PBRange)                 ' Set progressbar range 0-100
    UpdateWindow(hWnd)
    Result=True
END FUNCTION

SUB UpdateGauge(barVal AS integer)
    IF barVal=KILL_PROGBAR THEN
        DestroyWindow(hProgBar)
    ELSE
        IF barVal>100 THEN barVal=100
        IF barVal<0 THEN barVal=0
        SendMessage(hProgBar,PBM_SETPOS,barVal,0)                   ' Set progress bar value
    END IF
    UpdateWindow(hParent)
    DoEvents
END SUB

$IFDEF TIME_ME
' tStart=0: Initial val returned, tStart>0: elapsed time returned
FUNCTION ElapsedTimer(tStart)
    Result=Timer-tStart
END FUNCTION
$ENDIF
